'''
Date: 2023-03-12 19:43:45
LastEditors: ThetisEliza wxf199601@gmail.com
LastEditTime: 2023-03-12 20:12:30
FilePath: /outlier/src/outlier/biz/events.py
'''
from enum import Enum

class Ops(Enum):
    Add = 0
    Rmv = 1
    Rcv = 2
